"""
weitersager.httpreceiver
~~~~~~~~~~~~~~~~~~~~~~~~

HTTP server to receive messages

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from dataclasses import dataclass
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
import json
import sys
from typing import Optional, Set

from .signals import message_received
from .util import log, start_thread


@dataclass(frozen=True)
class Config:
    """An HTTP receiver configuration."""
    host: str
    port: int
    api_tokens: Optional[Set[str]] = None


@dataclass(frozen=True)
class Message:
    channel: str
    text: str


def parse_json_message(json_data):
    """Extract message from JSON."""
    data = json.loads(json_data)

    channel = data['channel']
    text = data['text']

    return Message(channel=channel, text=text)


class RequestHandler(BaseHTTPRequestHandler):
    """Handler for messages submitted via HTTP."""

    def do_POST(self):
        valid_api_tokens = self.server.api_tokens
        if valid_api_tokens:
            api_token = self._get_api_token()
            if not api_token:
                self.send_response(HTTPStatus.UNAUTHORIZED)
                self.end_headers()
                return

            if api_token not in valid_api_tokens:
                self.send_response(HTTPStatus.FORBIDDEN)
                self.end_headers()
                return

        try:
            content_length = int(self.headers.get('Content-Length', 0))
            data = self.rfile.read(content_length).decode('utf-8')
            message = parse_json_message(data)
        except (KeyError, ValueError):
            log(f'Invalid message received from {self.address_string()}.')
            self.send_error(HTTPStatus.BAD_REQUEST)
            return

        self.send_response(HTTPStatus.ACCEPTED)
        self.end_headers()

        message_received.send(
            channel_name=message.channel,
            text=message.text,
            source_address=self.client_address,
        )

    def _get_api_token(self):
        authorization_value = self.headers.get('authorization')
        if not authorization_value:
            return None

        prefix = 'WTRSGR '
        if not authorization_value.startswith(prefix):
            return None

        return authorization_value[len(prefix) :]

    def version_string(self):
        """Return custom server version string."""
        return 'Weitersager'


class ReceiveServer(HTTPServer):
    """HTTP server that waits for messages."""

    def __init__(self, config):
        address = (config.host, config.port)
        HTTPServer.__init__(self, address, RequestHandler)
        log('Listening for HTTP requests on {}:{:d}.', *address)

        self.api_tokens = config.api_tokens

def start_receive_server(config):
    """Start in a separate thread."""
    try:
        receiver = ReceiveServer(config)
    except Exception as e:
        sys.stderr.write(f'Error {e.errno:d}: {e.strerror}\n')
        sys.stderr.write(
            f'Probably no permission to open port {config.port}. '
            'Try to specify a port number above 1,024 (or even '
            '4,096) and up to 65,535.\n'
        )
        sys.exit(1)

    thread_name = receiver.__class__.__name__
    start_thread(receiver.serve_forever, thread_name)



"""
weitersager.irc
~~~~~~~~~~~~~~~

Internet Relay Chat

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from dataclasses import dataclass
from typing import List, Optional

from irc.bot import ServerSpec, SingleServerIRCBot
from jaraco.stream.buffer import LenientDecodingLineBuffer

from .signals import channel_joined, shutdown_requested
from .util import log, start_thread


@dataclass(frozen=True)
class Server:
    """An IRC server."""
    host: str
    port: int
    password: Optional[str] = None


@dataclass(frozen=True)
class Channel:
    """An IRC channel with optional password."""
    name: str
    password: Optional[str] = None


@dataclass(frozen=True)
class Config:
    server: Optional[Server]
    nickname: str
    realname: str
    channels: List[Channel]


class Bot(SingleServerIRCBot):
    """An IRC bot to forward messages to IRC channels."""

    def __init__(
        self,
        server,
        nickname,
        realname,
        channels,
        *,
        shutdown_predicate=None,
    ):
        log('Connecting to IRC server {0.host}:{0.port:d} ...', server)

        server_spec = ServerSpec(server.host, server.port, server.password)
        SingleServerIRCBot.__init__(self, [server_spec], nickname, realname)

        # Avoid `UnicodeDecodeError` on non-UTF-8 messages.
        self.connection.buffer_class = LenientDecodingLineBuffer

        # Note: `self.channels` already exists in super class.
        self.channels_to_join = channels

        self.shutdown_predicate = shutdown_predicate

    def start(self):
        """Connect to the server, in a separate thread."""
        start_thread(super().start, self.__class__.__name__)

    def get_version(self):
        return 'Weitersager'

    def on_welcome(self, conn, event):
        """Join channels after connect."""
        log('Connected to {}:{:d}.', *conn.socket.getpeername())

        channel_names = sorted(c.name for c in self.channels_to_join)
        log('Channels to join: {}', ', '.join(channel_names))

        for channel in self.channels_to_join:
            log('Joining channel {} ...', channel.name)
            conn.join(channel.name, channel.password or '')

    def on_nicknameinuse(self, conn, event):
        """Choose another nickname if conflicting."""
        self._nickname += '_'
        conn.nick(self._nickname)

    def on_join(self, conn, event):
        """Successfully joined channel."""
        joined_nick = event.source.nick
        channel_name = event.target

        if joined_nick == self._nickname:
            log('Joined IRC channel: {}', channel_name)
            channel_joined.send(channel_name=channel_name)

    def on_badchannelkey(self, conn, event):
        """Channel could not be joined due to wrong password."""
        channel = event.arguments[0]
        log('Cannot join channel {} (bad key).', channel)

    def on_privmsg(self, conn, event):
        """React on private messages."""
        nickmask = event.source
        text = event.arguments[0]
        if self.shutdown_predicate and self.shutdown_predicate(nickmask, text):
            self.shutdown(nickmask)

    def shutdown(self, nickmask):
        """Shut the bot down."""
        log('Shutdown requested by {}.', nickmask)
        shutdown_requested.send()
        self.die('Shutting down.')  # Joins IRC bot thread.

    def say(self, sender, *, channel_name=None, text=None):
        """Say message on channel."""
        self.connection.privmsg(channel_name, text)


class DummyBot:

    def __init__(self, server, nickname, realname, channels, **options):
        self.channels = channels

    def start(self):
        # Fake channel joins.
        for channel in self.channels:
            channel_joined.send(channel_name=channel.name)

    def say(self, sender, *, channel_name=None, text=None):
        log('{}> {}', channel_name, text)


def create_bot(config, **options):
    """Create and return an IRC bot according to the configuration."""
    if config.server:
        bot_class = Bot
    else:
        log('No IRC server specified; will write to STDOUT instead.')
        bot_class = DummyBot

    return bot_class(
        config.server,
        config.nickname,
        config.realname,
        config.channels,
        **options,
    )


def default_shutdown_predicate(nickmask, text):
    """Determine if this is a valid shutdown request."""
    return text == 'shutdown!'



"""
weitersager.config
~~~~~~~~~~~~~~~~~~

Configuration loading

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import rtoml

from .httpreceiver import Config as HttpConfig
from .irc import Channel, Config as IrcConfig, Server as IrcServer


DEFAULT_HTTP_HOST = '127.0.0.1'
DEFAULT_HTTP_PORT = 8080
DEFAULT_IRC_SERVER_PORT = 6667
DEFAULT_IRC_REALNAME = 'Weitersager'


def load_config(path):
    """Load configuration from file."""
    data = rtoml.load(path)

    http_config = _get_http_config(data)
    irc_config = _get_irc_config(data)

    return irc_config, http_config


def _get_http_config(data):
    data_http = data.get('http', {})

    host = data_http.get('host', DEFAULT_HTTP_HOST)
    port = int(data_http.get('port', DEFAULT_HTTP_PORT))
    api_tokens = data_http.get('api_tokens')
    if api_tokens:
        api_tokens = set(api_tokens)

    return HttpConfig(host, port, api_tokens)


def _get_irc_config(data):
    data_irc = data['irc']

    server = _get_irc_server(data_irc)
    nickname = data_irc['bot']['nickname']
    realname = data_irc['bot'].get('realname', DEFAULT_IRC_REALNAME)
    channels = list(_get_irc_channels(data_irc))

    return IrcConfig(
        server=server,
        nickname=nickname,
        realname=realname,
        channels=channels,
    )


def _get_irc_server(data_irc):
    data_server = data_irc.get('server')
    if data_server is None:
        return None

    host = data_server.get('host')
    if not host:
        return None

    port = int(data_server.get('port', DEFAULT_IRC_SERVER_PORT))
    password = data_server.get('password')

    return IrcServer(host, port, password)


def _get_irc_channels(data_irc):
    for channel in data_irc.get('channels', []):
        name = channel['name']
        password = channel.get('password')
        yield Channel(name, password)



"""
weitersager.cli
~~~~~~~~~~~~~~~

Command line entry point

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from argparse import ArgumentParser
from pathlib import Path

from .config import load_config
from .processor import start


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(prog='weitersager')
    parser.add_argument('config_filename', type=Path)
    return parser.parse_args()


def main():
    """Load the configuration file, start the IRC bot and HTTP listen server."""
    args = parse_args()

    irc_config, http_config = load_config(args.config_filename)

    start(irc_config, http_config)


if __name__ == '__main__':
    main()



"""
weitersager.processor
~~~~~~~~~~~~~~~~~~~~~

Connect HTTP server and IRC bot.

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from time import sleep

from .httpreceiver import start_receive_server
from .irc import create_bot
from .signals import (
    channel_joined,
    message_approved,
    message_received,
    shutdown_requested,
)
from .util import log


class Processor:

    def __init__(self):
        self.enabled_channel_names = set()
        self.shutdown = False

    def connect_to_signals(self):
        channel_joined.connect(self.enable_channel)
        message_received.connect(self.handle_message)
        shutdown_requested.connect(self.handle_shutdown_requested)

    def enable_channel(self, sender, *, channel_name=None):
        log('Enabled forwarding to channel {}.', channel_name)
        self.enabled_channel_names.add(channel_name)

    def handle_message(
        self, sender, *, channel_name=None, text=None, source_address=None
    ):
        """Log and announce an incoming message."""
        source = f'{source_address[0]}:{source_address[1]:d}'

        log(
            'Received message from {} for channel {} with text "{}"',
            source,
            channel_name,
            text,
        )

        if channel_name not in self.enabled_channel_names:
            log(
                'Could not send message to channel {}, not joined.',
                channel_name,
            )
            return

        message_approved.send(channel_name=channel_name, text=text)

    def handle_shutdown_requested(self, sender):
        self.shutdown = True

    def run(self):
        """Run the main loop until shutdown is requested."""
        while not self.shutdown:
            sleep(0.5)

        log('Shutting down ...')


def start(irc_config, http_config, **options):
    """Start the IRC bot and HTTP listen server."""
    bot = create_bot(irc_config, **options)
    message_approved.connect(bot.say)

    processor = Processor()

    # Up to this point, no signals must have been sent.

    processor.connect_to_signals()

    # Signals are allowed be sent from here on.

    start_receive_server(http_config)
    bot.start()

    processor.run()



"""
weitersager.util
~~~~~~~~~~~~~~~~

Utilities

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from datetime import datetime
from threading import Thread


def log(message, *args, **kwargs):
    timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print(timestamp, message.format(*args, **kwargs))


def start_thread(target, name):
    """Create, configure, and start a new thread."""
    t = Thread(target=target, name=name, daemon=True)
    t.start()



"""
weitersager.signals
~~~~~~~~~~~~~~~~~~~

Signals

:Copyright: 2007-2020 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from blinker import signal


channel_joined = signal('channel-joined')
message_received = signal('message-received')
message_approved = signal('message-approved')
shutdown_requested = signal('shutdown-requested')
